
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "sfnt.h"

uint32_t
read_uint32(unsigned char * data)
{
	return (data[0]<<24)|(data[1]<<16)|(data[2]<<8)|data[3] ;
}

uint16_t
read_uint16(unsigned char * data)
{
	return (data[0]<<8)|(data[1]) ;
}

void
write_uint32(unsigned char * data, uint32_t x)
{
	data[0] = (x&0x0ff000000)>>24 ;
	data[1] = (x&0x000ff0000)>>16 ;
	data[2] = (x&0x00000ff00)>> 8 ;
	data[3] = (x&0x0000000ff)     ;
}

void
write_uint16(unsigned char * data, uint16_t x)
{
	data[0] = (x&0x0ff00)>>8 ;
	data[1] = (x&0x000ff)    ;
}

uint32_t CalcTableChecksum(unsigned char * data, uint32_t numberOfBytesInTable)
{
	uint32_t i ;
    uint32_t sum = 0;
    uint32_t nLongs = (numberOfBytesInTable + 3) / 4;
	
	for (i=0; i<nLongs*4; i += 4) {
		sum += read_uint32(data+i) ;
	}
	
    return sum;
}

struct sfnt_face_info *
parse_sfnt_face_info(unsigned char * storage, uint32_t length)
{
	int j ;
	struct sfnt_face_info * finfo ;
	int rpos = 0 ;
	
	finfo = malloc(sizeof(struct sfnt_face_info)) ;
	
	finfo->header.sctype     = read_uint32(storage +  0) ;
	finfo->header.numtables  = read_uint16(storage +  4) ;
	finfo->header.srange     = read_uint16(storage +  6) ;
	finfo->header.entrysel   = read_uint16(storage +  8) ;
	finfo->header.rangeshift = read_uint16(storage + 10) ;
	
	rpos += 12 ;
	
	finfo->tables = malloc(sizeof(struct sfnt_table)*finfo->header.numtables) ;
	
	for (j=0; j<finfo->header.numtables; j++) {
		memcpy(finfo->tables[j].tag, storage + rpos, 4) ;
		finfo->tables[j].csum   = read_uint32(storage + rpos +  4) ;
		finfo->tables[j].offset = read_uint32(storage + rpos +  8) ;
		finfo->tables[j].length = read_uint32(storage + rpos + 12) ;
		finfo->tables[j].padded_length = ((finfo->tables[j].length+3)/4)*4 ;
		finfo->tables[j].entity = storage + finfo->tables[j].offset ;
		rpos += 16 ;
	}
	
	return finfo ;
}

struct name_and_count
parse_sfnt_name_table(unsigned char * storage, uint32_t length)
{
	struct name_and_count ret ;
	uint16_t format ;
	uint16_t count ;
	uint16_t stroffset ;
	uint16_t j ;
	
	struct sfnt_name_entry * names ;
	
	format    = read_uint16(storage + 0) ;
	count     = read_uint16(storage + 2) ;
	stroffset = read_uint16(storage + 4) ;
	
	names = calloc(count, sizeof(struct sfnt_name_entry)) ;
	ret.count = count ;
	ret.value = names ;
	
	for (j=0; j<count; j++) {
		uint16_t ofs ;
		
		names[j].platformID         = read_uint16(storage + 6+j*12 + 0) ;
		names[j].platformSpecificID = read_uint16(storage + 6+j*12 + 2) ;
		names[j].languageID         = read_uint16(storage + 6+j*12 + 4) ;
		names[j].nameID             = read_uint16(storage + 6+j*12 + 6) ;
		names[j].length             = read_uint16(storage + 6+j*12 + 8) ;
		
		names[j].entity = malloc(names[j].length + 2) ;
		ofs  = read_uint16(storage + 6+j*12 + 10) ;
		
		memcpy(names[j].entity, storage + ofs + stroffset, names[j].length) ;
		names[j].entity[names[j].length + 0] = 0 ;
		names[j].entity[names[j].length + 1] = 0 ;
	}
	
	return ret ;
}

struct uchar_and_length
render_sfnt_name_table(int count, struct sfnt_name_entry * names)
{
	int i ;
	uint16_t format = 0 ;
	uint16_t stroffset ;
	int length = 0 ;
	unsigned char * storage ;
	int wpos ;
	struct uchar_and_length ret ;
	
	stroffset = 6 + 12*count ;
	
	for (i=0; i<count; i++) {
		names[i].offset = length ;
		length += names[i].length ;
	}
	
	storage = malloc(((6 + 12*count + length)+3)/4*4) ;
	wpos = 0 ;
	
	write_uint16(storage +  0, format) ;
	write_uint16(storage +  2, count) ;
	write_uint16(storage +  4, stroffset) ;
	
	wpos += 6 ;
	
	for (i=0; i<count; i++) {
		write_uint16(storage + 6 + i*12 +  0, names[i].platformID) ;
		write_uint16(storage + 6 + i*12 +  2, names[i].platformSpecificID) ;
		write_uint16(storage + 6 + i*12 +  4, names[i].languageID) ;
		write_uint16(storage + 6 + i*12 +  6, names[i].nameID) ;
		write_uint16(storage + 6 + i*12 +  8, names[i].length) ;
		write_uint16(storage + 6 + i*12 + 10, names[i].offset) ;
		
		memcpy(storage + stroffset + names[i].offset, names[i].entity, names[i].length) ;
	}
	
	ret.value = storage ;
	ret.length = 6 + 12*count + length ;
	return ret ;
}


struct sfnt_name_entry *
find_sfnt_name_entry(int count, struct sfnt_name_entry * names,
					 uint16_t platformID, uint16_t platformSpecificID, uint16_t languageID, uint16_t nameID)
{
	int i ;
	
	for (i=0; i<count; i++) {
		if (names[i].platformID == platformID &&
			names[i].platformSpecificID == platformSpecificID &&
			names[i].languageID == languageID &&
			names[i].nameID == nameID) {
			return names + i ;
		}
	}
	
	return NULL ;
}

struct sfnt_cmap_table *
parse_sfnt_cmap_table(unsigned char * storage, int length)
{
	struct sfnt_cmap_table * cmap ;
	uint16_t i ;
	uint32_t total = 0 ;
	
	cmap = malloc(sizeof(struct sfnt_cmap_table)) ;
	
	cmap->version    = read_uint16(storage + 0) ;
	cmap->numberSubtables = read_uint16(storage + 2) ;
	
	cmap->subtables  = malloc(sizeof(struct sfnt_cmap_subtable)*cmap->numberSubtables) ;
	
	total = 4 + cmap->numberSubtables * 8 ;
	
	for (i=0; i<cmap->numberSubtables; i++) {
		struct sfnt_cmap_subtable * stbl = cmap->subtables + i ;
		
		stbl->platformID = read_uint16(storage + 4 + i*8 + 0) ;
		stbl->platformSpecificID = read_uint16(storage + 4 + i*8 + 2) ;
		stbl->offset = read_uint32(storage + 4 + i*8 + 4) ;
		stbl->format = read_uint16(storage + stbl->offset + 0) ;
		
		switch (stbl->format) {
		case 0: case 2: case 4: case 6:
			stbl->length = read_uint16(storage + stbl->offset + 2) ;
			break ;
			
		case 8:
			stbl->length = read_uint32(storage + stbl->offset + 2) ;
			break ;
			
		default:
			stbl->length = 0 ;
		}
		
		total += stbl->length ;
	}
	
	if (length - total > 0) {
		printf("borken cmap table\n") ;
	}
	
	for (i=0; i<cmap->numberSubtables; i++) {
		struct sfnt_cmap_subtable * stbl = cmap->subtables + i ;
		
		if (stbl->platformID == TT_PLATFORM_MICROSOFT &&
			stbl->platformSpecificID == TT_MS_ID_UNICODE_CS &&
			stbl->format == 4 &&
			(length - total) > 0) {
			stbl->length += length - total ;
			total = length ;
		}
		
		if (stbl->length > 0) {
			stbl->entity = malloc(stbl->length) ;
			memcpy(stbl->entity, storage + stbl->offset, stbl->length) ;
		} else {
			stbl->entity = NULL ;
		}
	}
	
	return cmap ;
}

struct uchar_and_length
render_sfnt_cmap_table(struct sfnt_cmap_table * cmap)
{
	int i ;
	uint32_t length = 4 ;
	unsigned char * storage ;
	struct uchar_and_length ret ;
	
	length += 8*cmap->numberSubtables ;
	
	for (i=0; i<cmap->numberSubtables; i++) {
		struct sfnt_cmap_subtable * stbl = cmap->subtables + i ;
		stbl->offset = length ;
		length += stbl->length ;
	}
	
	length += 4 ;
	storage = malloc(((length+3)/4)*4) ;
	
	write_uint16(storage + 0, cmap->version) ;
	write_uint16(storage + 2, cmap->numberSubtables) ;
	
	for (i=0; i<cmap->numberSubtables; i++) {
		struct sfnt_cmap_subtable * stbl = cmap->subtables + i ;
		write_uint16(storage + 4 + i*8 + 0, stbl->platformID) ;
		write_uint16(storage + 4 + i*8 + 2, stbl->platformSpecificID) ;
		write_uint32(storage + 4 + i*8 + 4, stbl->offset) ;
		
		memcpy(storage + stbl->offset, stbl->entity, stbl->length) ;
	}
	
	ret.length = length ;
	ret.value = storage ;
	return ret ;
}

int
write_sfnt_face(const char * filename, struct sfnt_face_info * finfo)
{
	FILE * fp ;
	uint32_t tblcnt ;
	uint32_t i ;
	uint32_t headersize ;
	uint32_t len ;
	unsigned char * hdrbuf ;
	
	fp = fopen(filename, "w") ;
	if (fp == NULL) {
		perror("fopen") ;
		return -1 ;
	}
	
	tblcnt = finfo->header.numtables ;
	fprintf(stderr, "%d tables\n", tblcnt) ;
	
	headersize = 12 + 16*tblcnt ;
	
	len = 0 ;
	for (i=0; i<tblcnt; i++) {
		struct sfnt_table * table = finfo->tables + i ;
		
		table->offset = headersize + len ;
		len += table->padded_length ;
		
		table->csum = CalcTableChecksum(table->entity, table->length) ;
	}
	
	hdrbuf = malloc(headersize) ;
	write_uint32(hdrbuf +  0, finfo->header.sctype) ;
	write_uint16(hdrbuf +  4, finfo->header.numtables) ;
	write_uint16(hdrbuf +  6, finfo->header.srange) ;
	write_uint16(hdrbuf +  8, finfo->header.entrysel) ;
	write_uint16(hdrbuf + 10, finfo->header.rangeshift) ;
	
	for (i=0; i<tblcnt; i++) {
		struct sfnt_table * table = finfo->tables + i ;
		
		memcpy(hdrbuf + 12 + i*16, table->tag, 4) ;
		write_uint32(hdrbuf + 12 + i*16 +  4, table->csum) ;
		write_uint32(hdrbuf + 12 + i*16 +  8, table->offset) ;
		write_uint32(hdrbuf + 12 + i*16 + 12, table->length) ;
	}
	
	fwrite(hdrbuf, 1, headersize, fp) ;
	free(hdrbuf) ;
	
	for (i=0; i<tblcnt; i++) {
		fwrite(finfo->tables[i].entity, 1, finfo->tables[i].padded_length, fp) ;
	}
	
	fclose(fp) ;
	return 0 ;
}
