/* -*- c -*- */

#ifndef SFNT_H
#define SFNT_H

#include <stdint.h>

typedef unsigned char sfnt_tag_t[4] ;

struct sfnt_header {
	uint32_t sctype ;
	uint16_t numtables ;
	uint16_t srange ;
	uint16_t entrysel ;
	uint16_t rangeshift ;
} ;

struct sfnt_face_info {
	uint32_t offset ;
	struct sfnt_header header ;
	struct sfnt_table * tables ;
} ;

struct sfnt_table {
	sfnt_tag_t tag ;
	uint32_t csum ;
	uint32_t offset ;
	uint32_t length ;
	uint32_t padded_length ;
	unsigned char * entity ;
} ;

struct sfnt_cmap_table {
	uint16_t version ;
	uint16_t numberSubtables ;
	struct sfnt_cmap_subtable * subtables ;
} ;

struct sfnt_cmap_subtable {
	uint16_t platformID ;
	uint16_t platformSpecificID ;
	uint32_t offset ;
	uint16_t format ;
	uint32_t length ;
	unsigned char * entity ;
} ;

struct sfnt_name_entry {
	uint16_t platformID ;
	uint16_t platformSpecificID ;
	uint16_t languageID ;
	uint16_t nameID ;
	uint16_t length ;
	uint16_t offset ;
	unsigned char * entity ;
} ;

struct name_and_count {
	struct sfnt_name_entry * value ;
	int count ;
} ;

struct uchar_and_length {
	unsigned char * value ;
	int length ;
} ;


/* -------------- divert from FreeType ----------------- */

#define TT_PLATFORM_APPLE_UNICODE  0
#define TT_PLATFORM_MACINTOSH      1
#define TT_PLATFORM_ISO            2 /* deprecated */
#define TT_PLATFORM_MICROSOFT      3
#define TT_PLATFORM_CUSTOM         4
#define TT_PLATFORM_ADOBE          7 /* artificial */

#define TT_APPLE_ID_DEFAULT      0 /* Unicode 1.0 */
#define TT_APPLE_ID_UNICODE_1_1  1 /* specify Hangul at U+34xx */
#define TT_APPLE_ID_ISO_10646    2 /* deprecated */
#define TT_APPLE_ID_UNICODE_2_0  3 /* or later */
#define TT_APPLE_ID_UNICODE_32   4 /* 2.0 or later, full repertoire */

#define TT_MAC_ID_ROMAN                 0
#define TT_MAC_ID_JAPANESE              1
#define TT_MAC_ID_TRADITIONAL_CHINESE   2
#define TT_MAC_ID_KOREAN                3
#define TT_MAC_ID_ARABIC                4
#define TT_MAC_ID_HEBREW                5
#define TT_MAC_ID_GREEK                 6
#define TT_MAC_ID_RUSSIAN               7
#define TT_MAC_ID_RSYMBOL               8
#define TT_MAC_ID_DEVANAGARI            9
#define TT_MAC_ID_GURMUKHI             10
#define TT_MAC_ID_GUJARATI             11
#define TT_MAC_ID_ORIYA                12
#define TT_MAC_ID_BENGALI              13
#define TT_MAC_ID_TAMIL                14
#define TT_MAC_ID_TELUGU               15
#define TT_MAC_ID_KANNADA              16
#define TT_MAC_ID_MALAYALAM            17
#define TT_MAC_ID_SINHALESE            18
#define TT_MAC_ID_BURMESE              19
#define TT_MAC_ID_KHMER                20
#define TT_MAC_ID_THAI                 21
#define TT_MAC_ID_LAOTIAN              22
#define TT_MAC_ID_GEORGIAN             23
#define TT_MAC_ID_ARMENIAN             24
#define TT_MAC_ID_MALDIVIAN            25
#define TT_MAC_ID_SIMPLIFIED_CHINESE   25
#define TT_MAC_ID_TIBETAN              26
#define TT_MAC_ID_MONGOLIAN            27
#define TT_MAC_ID_GEEZ                 28
#define TT_MAC_ID_SLAVIC               29
#define TT_MAC_ID_VIETNAMESE           30
#define TT_MAC_ID_SINDHI               31
#define TT_MAC_ID_UNINTERP             32

#define TT_MS_ID_SYMBOL_CS    0
#define TT_MS_ID_UNICODE_CS   1
#define TT_MS_ID_SJIS         2
#define TT_MS_ID_GB2312       3
#define TT_MS_ID_BIG_5        4
#define TT_MS_ID_WANSUNG      5
#define TT_MS_ID_JOHAB        6
#define TT_MS_ID_UCS_4       10

#define TT_MAC_LANGID_ENGLISH                       0
#define TT_MAC_LANGID_JAPANESE                     11

#define TT_MS_LANGID_ENGLISH_UNITED_STATES             0x0409
#define TT_MS_LANGID_JAPANESE_JAPAN                    0x0411

#define TT_NAME_ID_FONT_FAMILY          1
#define TT_NAME_ID_FONT_SUBFAMILY       2
#define TT_NAME_ID_UNIQUE_ID            3
#define TT_NAME_ID_FULL_NAME            4
#define TT_NAME_ID_VERSION_STRING       5
#define TT_NAME_ID_PS_NAME              6
#define TT_NAME_ID_TRADEMARK            7

  /* the following values are from the OpenType spec */
#define TT_NAME_ID_MANUFACTURER         8
#define TT_NAME_ID_DESIGNER             9
#define TT_NAME_ID_DESCRIPTION          10
#define TT_NAME_ID_VENDOR_URL           11
#define TT_NAME_ID_DESIGNER_URL         12
#define TT_NAME_ID_LICENSE              13
#define TT_NAME_ID_LICENSE_URL          14
  /* number 15 is reserved */
#define TT_NAME_ID_PREFERRED_FAMILY     16
#define TT_NAME_ID_PREFERRED_SUBFAMILY  17
#define TT_NAME_ID_MAC_FULL_NAME        18

  /* The following code is new as of 2000-01-21 */
#define TT_NAME_ID_SAMPLE_TEXT          19

  /* This is new in OpenType 1.3 */
#define TT_NAME_ID_CID_FINDFONT_NAME    20

/* ----------------------------------------------------- */


uint32_t read_uint32(unsigned char *) ;
uint16_t read_uint16(unsigned char *) ;
void write_uint32(unsigned char *, uint32_t) ;
void write_uint16(unsigned char *, uint16_t) ;
uint32_t CalcTableChecksum(unsigned char *, uint32_t) ;

struct sfnt_face_info * parse_sfnt_face_info(unsigned char *, uint32_t) ;
struct name_and_count parse_sfnt_name_table(unsigned char *, uint32_t) ;
struct uchar_and_length render_sfnt_name_table(int, struct sfnt_name_entry *) ;

struct sfnt_name_entry *
find_sfnt_name_entry(int count,
					 struct sfnt_name_entry * names,
					 uint16_t platformID,
					 uint16_t platformSpecificID,
					 uint16_t languageID,
					 uint16_t nameID) ;

struct sfnt_cmap_table * parse_sfnt_cmap_table(unsigned char *, int) ;
struct uchar_and_length render_sfnt_cmap_table(struct sfnt_cmap_table *) ;

int write_sfnt_face(const char *, struct sfnt_face_info *) ;


#endif

