
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ttc.h"

struct lang_entry {
	uint16_t id ;
	char * str ;
} ;

struct cmap4_entry {
	uint16_t startCode ;
	uint16_t endCode ;
	uint16_t idDelta ;
	uint16_t idRangeOffset ;
	uint16_t glyfOffset ;
} ;

char * platform_strings[] = {
	"Unicode",
	"Macintosh",
	"(reserved)",
	"Microsoft",
	"unknown(4)",
	"unknown(5)",
	"unknown(6)",
	"unknown(7)",
	"unknown(8)",
	"unknown(9)",
	"unknown(10)",
	"unknown(11)",
	"unknown(12)",
	"unknown(13)",
	"unknown(14)",
	"unknown(15)",
} ;

char * platform_mac_specs[] = {
	"Roman",      /* 0 */
	"Japanese",   /* 1 */
	"Traditional Chinese",
	"Korean",
	"Arabic",
	"Hebrew",     /* 5 */
	"Greek",
	"Russian",
	"RSymbol",
	"Devanagari",
	"Gurmukhi",   /* 10 */
	"Gujarati",
	"Oriya",
	"Bengali",
	"Tamil",
	"Telugu",     /* 15 */
	"Kannada",
	"Malayalam",
	"Sinhalese",
	"Burmese",
	"Khmer",      /* 20 */
	"Thai",
	"Laotian",
	"Georgian",
	"Armenian",
	"Simplified Chinese", /* 25 */
	"Tibetan",
	"Mongolian",
	"Geez",
	"Slavic",
	"Vietnamese", /* 30 */
	"Sindhi",
	"(Uninterpreted)", /* 32 */
} ;

char * platform_ms_specs[] = {
	"Symbol",
	"Unicode",
	"SJIS",
	"GB2312",
	"Big-5",
	"Wansung",
	"Johab",
	"UCS-4",
} ;

char * nameid_string[] = {
	"Copyright",
	"Font Family",
	"Font Subfamily",
	"Unique ID",
	"Full Name",
	"Version String",
	"PostScript Name",
	"Trademark",
} ;

struct lang_entry lang_ms_string[] = {
	{0x0409, "en_US"},
	{0x0411, "ja_JP"},
	{0xffff, NULL},
} ;

struct lang_entry lang_mac_string[] = {
	{0x0000, "English"},
	{0x000b, "Japanese"},
	{0xffff, NULL},
} ;

void
read_sfnt_name_table(int length, unsigned char * storage)
{
	uint16_t format ;
	uint16_t count ;
	uint16_t stroffset ;
	uint16_t j ;
	
	format = (storage[0]<<8) | (storage[1]) ;
	count = (storage[2]<<8) | (storage[3]) ;
	stroffset = (storage[4]<<8) | (storage[5]) ;
	
	printf("    format: %u\n", format) ;
	printf("    count : %u\n", count) ;
	printf("    stringOffset: %u\n", stroffset) ;
	printf("\n") ;
	
	for (j=0; j<count; j++) {
		uint16_t k ;
		uint16_t pid, psid, lid, nid, l, ofs ;
		pid  = (storage[6+j*12+ 0]<<8) | (storage[6+j*12+ 1]) ;
		psid = (storage[6+j*12+ 2]<<8) | (storage[6+j*12+ 3]) ;
		lid  = (storage[6+j*12+ 4]<<8) | (storage[6+j*12+ 5]) ;
		nid  = (storage[6+j*12+ 6]<<8) | (storage[6+j*12+ 7]) ;
		l    = (storage[6+j*12+ 8]<<8) | (storage[6+j*12+ 9]) ;
		ofs  = (storage[6+j*12+10]<<8) | (storage[6+j*12+11]) ;
		
		printf("    platform: %s, ", platform_strings[pid]) ;
		switch (pid) {
		case 0: /* Unicode */
			printf("%d\n", psid) ;
			break ;
		case 1: /* Macintosh */
			printf("%s\n", platform_mac_specs[psid]) ;
			break ;
		case 2: /* never reach ? */
			printf("\n") ;
			break ;
		case 3: /* Microsoft */
			printf("%s\n", platform_ms_specs[psid]) ;
			break ;
		}
		
		switch (pid) {
		case 0: /* Unicode */
			printf("    language: %d\n", lid) ;
			break ;
			
		case 1: /* Macintosh */
			k = 0 ;
			while (1) {
				if (lang_mac_string[k].id == 0x0ffff) {
					printf("    language: %d\n", lid) ;
					break ;
				} else if (lang_mac_string[k].id == lid) {
					printf("    language: %s\n", lang_mac_string[k].str) ;
					break ;
				}
				
				k++ ;
			}
			break ;
			
		case 2: /* never reach ? */
			printf("\n") ;
			break ;
			
		case 3: /* Microsoft */
			k = 0 ;
			while (1) {
				if (lang_ms_string[k].id == 0x0ffff) {
					printf("    language: %d\n", lid) ;
					break ;
				} else if (lang_ms_string[k].id == lid) {
					printf("    language: %s\n", lang_ms_string[k].str) ;
					break ;
				}
				
				k++ ;
			}
			break ;
		}
		
		printf("    nameID: %s\n", nameid_string[nid]) ;
		if (pid == 1 && psid == 0) {
			printf("    ") ;
			for (k=0; k<l; k++) {
				printf("%c", storage[k+ofs+stroffset]) ;
			}
			printf("\n") ;
		}
		
		printf("\n") ;
	}
}

void
show_cmap_format6(unsigned char * data, uint16_t length)
{
	uint16_t i ;
	uint16_t language ;
	uint16_t firstCode ;
	uint16_t entryCount ;
	
	language   = (data[4 +  0]<<8) | (data[4 +  1]) ;
	firstCode  = (data[4 +  2]<<8) | (data[4 +  3]) ;
	entryCount = (data[4 +  4]<<8) | (data[4 +  5]) ;
	
	printf("     language  : %u\n", language) ;
	printf("     firstCode : %04x\n", firstCode) ;
	printf("     entryCount: %u\n", entryCount) ;
	
	printf("     -- mapping\n") ;
	for (i=0; i<entryCount; i++) {
		uint16_t tocode = read_uint16(data + 10 + i*2) ;
		if (tocode) 
			printf("      %04x -> %04x\n", i + firstCode, tocode) ;
	}
}

void
show_cmap_format4(unsigned char * data, uint16_t length)
{
	uint16_t i ;
	uint16_t language, segCount, searchRange, entrySelector, rangeShift ;
	struct cmap4_entry * segs ;
	
	language    = (data[4 +  0]<<8) | (data[4 +  1]) ;
	segCount  = ((data[4 +  2]<<8) | (data[4 +  3]))/2 ;
	searchRange = (data[4 +  4]<<8) | (data[4 +  5]) ;
	entrySelector = (data[4 +  6]<<8) | (data[4 +  7]) ;
	rangeShift  = (data[4 +  8]<<8) | (data[4 +  9]) ;
	
	printf("     language: %u\n", language) ;
	printf("     segCount: %u\n", segCount) ;
	printf("     searchRange: %u\n", searchRange) ;	
	printf("     entrySelector: %u\n", entrySelector) ;	
	printf("     rangeShift: %u\n", rangeShift) ;
	
	printf("     -- segments\n") ;
	segs = malloc(sizeof(struct cmap4_entry)*segCount) ;
	for (i=0; i<segCount; i++) {
		uint16_t c ;
		
		segs[i].endCode = (data[14 + i*2 + 0]<<8)|data[14 + i*2 + 1] ;
		segs[i].startCode = (data[14+segCount*2+2+i*2]<<8)|data[14+segCount*2+2+i*2+1] ;
		segs[i].idDelta = (data[14+segCount*2*2+2+i*2]<<8)|data[14+segCount*2*2+2+i*2+1] ;
		segs[i].idRangeOffset = (data[14+segCount*2*3+2+i*2]<<8)|data[14+segCount*2*3+2+i*2+1] ;
		if (segs[i].idRangeOffset == 0) {
			segs[i].glyfOffset = 0 ;
		} else {
			segs[i].glyfOffset = 14+segCount*2*3+2+i*2 + segs[i].idRangeOffset ;
		}
		
		if (segs[i].endCode == 0x0ffff) break ;
		
		printf("     segment %u\n", i) ;
		printf("       start : %04x\n", segs[i].startCode) ;
		printf("       end   : %04x\n", segs[i].endCode) ;
		printf("       delta : %u\n", segs[i].idDelta) ;
		printf("       offset: %04x\n", segs[i].glyfOffset) ;
		
		if (segs[i].glyfOffset) {
			for (c = segs[i].startCode; c <= segs[i].endCode; c++) {
				uint16_t g = read_uint16(data + segs[i].glyfOffset + (c - segs[i].startCode)*2) ;
				if (g) printf("       %04x -> %04x\n", c, g) ;
			}
		} else {
			for (c = segs[i].startCode; c <= segs[i].endCode; c++) {
				printf("       %04x -> %04x\n", c, c+segs[i].idDelta) ;
			}
		}
	}
	free(segs) ;
}

void
read_sfnt_cmap_table(int length, unsigned char * storage)
{
	uint16_t version ;
	uint16_t numSubTbls ;
	uint16_t i ;
	
	version = (storage[0]<<8) | storage[1] ;
	numSubTbls = (storage[2]<<8) | storage[3] ;
	
	printf("    -- cmap index \n") ;
	printf("    version: %04x\n", version) ;
	printf("    number of subtables: %u\n", numSubTbls) ;
	
	printf("\n") ;
	
	printf("    -- cmap encoding subtables \n") ;
	for (i=0; i<numSubTbls; i++) {
		uint16_t pid, psid ;
		uint32_t ofs ;
		
		uint16_t format ;
		
		pid  = (storage[4 + i*8 + 0] << 8) | storage[4 + i*8 + 1] ;
		psid = (storage[4 + i*8 + 2] << 8) | storage[4 + i*8 + 3] ;
		ofs  = (storage[4 + i*8 + 4] << 24) | (storage[4 + i*8 + 5] << 16) |
			(storage[4 + i*8 + 6] << 8) | (storage[4 + i*8 + 7]) ;
		
		printf("     platform: %s, ", platform_strings[pid]) ;
		switch (pid) {
		case 0: /* Unicode */
			printf("%d\n", psid) ;
			break ;
		case 1: /* Macintosh */
			printf("%s\n", platform_mac_specs[psid]) ;
			break ;
		case 2: /* never reach ? */
			printf("\n") ;
			break ;
		case 3: /* Microsoft */
			printf("%s\n", platform_ms_specs[psid]) ;
			break ;
		}
		
		printf("     offset: %08x\n", ofs) ;
		
		format = (storage[ofs + 0]<<8) | storage[ofs + 1] ;
		printf("     format: %u\n", format) ;
		
		switch (format) {
		case 0:
		{
			break ;
		}
		
		case 2:
		{
			break ;
		}
		
		case 4:
		{
			uint16_t tlen = (storage[ofs + 2]<<8) | storage[ofs + 3] ;
			printf("     length: %u\n", tlen) ;
			show_cmap_format4(storage + ofs, tlen) ;
			break ;
		}
		
		case 6:
		{
			uint16_t tlen = (storage[ofs + 2]<<8) | storage[ofs + 3] ;
			printf("     length: %u\n", tlen) ;
			show_cmap_format6(storage + ofs, tlen) ;
			break ;
		}
		
		case 8:
		{
			break ;
		}
		}
		
		printf("\n") ;
	}
}


int
main(int argc, char ** argv)
{
	int i, j ;
	int ret ;
	int fd ;
	uint32_t length ;
	struct stat st ;
	unsigned char * data ;
	struct ttc_info * ttcf ;
	
	if (argc < 2) exit(1) ;
	
	fd = open(argv[1], O_RDONLY) ;
	if (fd < 0) {
		perror("open") ;
		exit(1) ;
	}
	
	ret = fstat(fd, &st) ;
	if (ret < 0) {
		perror("fstat") ;
		exit(1) ;
	}
	
	length = st.st_size ;
	data = mmap(0, length, PROT_READ, MAP_PRIVATE, fd, 0) ;
	if (data == MAP_FAILED) {
		perror("mmap") ;
		exit(1) ;
	}
	
	close(fd) ;
	
	ttcf = ttc_parse_binary(data, length) ;
	if (ttcf == NULL) {
		fprintf(stderr, "parse failed...\n") ;
		exit(1) ;
	}
	munmap(data, length) ;
	
	printf("%d faces\n", ttcf->face_count) ;
	
	for (i=0; i<ttcf->face_count; i++) {
		struct sfnt_face_info * face = ttcf->faces + i ;
		
		printf("face %d\n", i) ;
		
		printf("  offset: %08x\n", face->offset) ;
		printf("  numtables: %d\n", face->header.numtables) ;
		printf("  searchRange: %d\n", face->header.srange) ;
		printf("  entrySelector: %d\n", face->header.entrysel) ;
		printf("  rangeShift: %d\n", face->header.rangeshift) ;
		
		for (j=0; j<face->header.numtables; j++) {
			int idx = face->table_indexes[j] ;
			unsigned char buf[5] ;
			write_uint32(buf, read_uint32(ttcf->tables[idx].tag)) ;
			buf[4] = 0 ;
			printf("  tag: '%s'\n", (char *)buf) ;
			printf("    offset: %08x\n", ttcf->tables[idx].offset) ;
			printf("    length: %lu\n", (unsigned long)ttcf->tables[idx].length) ;
		}
		
		for (j=0; j<face->header.numtables; j++) {
			int idx = face->table_indexes[j] ;
			if (*((uint32_t *)ttcf->tables[idx].tag) == 0x656d616e) {
				printf("   NAME table\n") ;
				printf("    offset: %08x\n", ttcf->tables[idx].offset) ;
				read_sfnt_name_table(ttcf->tables[idx].length, ttcf->tables[idx].entity) ;
			} else if (*((uint32_t *)ttcf->tables[idx].tag) == 0x70616d63) {
				printf("   CMAP table\n") ;
				printf("    offset: %08x\n", ttcf->tables[idx].offset) ;
				read_sfnt_cmap_table(ttcf->tables[idx].length, ttcf->tables[idx].entity) ;
			}
		}
	}

	return 0 ;
}

