
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ttc.h"

int
load_ttf(struct ttc_info * ttcf, const char * filename)
{
	int i ;
	int ret ;
	int fd ;
	int idx ;
	uint32_t length ;
	struct stat st ;
	unsigned char * data ;
	struct sfnt_header header ;
	struct sfnt_face_info * face ;
	
	fd = open(filename, O_RDONLY) ;
	if (fd < 0) {
		perror("open") ;
		return -1 ;
	}
	
	ret = fstat(fd, &st) ;
	if (ret < 0) {
		perror("fstat") ;
		return -1 ;
	}
	
	length = st.st_size ;
	data = mmap(0, length, PROT_READ, MAP_PRIVATE, fd, 0) ;
	if (data == MAP_FAILED) {
		perror("mmap") ;
		close(fd) ;
		return -1 ;
	}
	
	close(fd) ;
	
	idx = ttc_create_face(ttcf) ;
	face = ttcf->faces + idx ;
	
	header.sctype     = read_uint32(data +  0) ;
	header.numtables  = read_uint16(data +  4) ;
	header.srange     = read_uint16(data +  6) ;
	header.entrysel   = read_uint16(data +  8) ;
	header.rangeshift = read_uint16(data + 10) ;
	
	face->header.sctype     = header.sctype ;
	face->header.srange     = header.srange ;
	face->header.entrysel   = header.entrysel ;
	face->header.rangeshift = header.rangeshift ;
	
	for (i=0; i<header.numtables; i++) {
		sfnt_tag_t tag ;
		uint32_t checksum ;
		uint32_t offset ;
		uint32_t length ;
		
		read_tag(tag, data + 12 + i*16 + 0) ;
		checksum = read_uint32(data + 12 + i*16 +  4) ;
		offset   = read_uint32(data + 12 + i*16 +  8) ;
		length   = read_uint32(data + 12 + i*16 + 12) ;
		
		ttc_add_table_to_face(ttcf, idx, tag, length, data + offset) ;
	}
	
	munmap(data, length) ;
	return 0 ;
}

int
main(int argc, char ** argv)
{
	FILE * fp ;
	int ret ;
	int i ;
	struct ttc_info * ttcf ;
	
	if (argc < 3) {
		fprintf(stderr, "usage: ttfjoin ttcfile ttffile [ttffile ...]\n") ;
		exit(1) ;
	}
	
	ttcf = create_ttc_info(0x00010000) ;
	
	for (i=2; i<argc; i++) {
		load_ttf(ttcf, argv[i]) ;
	}
	
	fp = fopen(argv[1], "w") ;
	if (fp == NULL) {
		perror("fopen") ;
		exit(2) ;
	}
	
	ret = ttc_write_to_file(ttcf, fp) ;
	if (ret < 0) {
		fprintf(stderr, "write failed.\n") ;
		exit(2) ;
	}
	fclose(fp) ;
	
	return 0 ;
}

